﻿const int Iterations = 10_000;

Test[] tests = [
	new(AddOnly, nameof(AddOnly)),
	new(ConsoleWriteLineAndAdd, nameof(ConsoleWriteLineAndAdd)),
	// new(StreamWriterWriteLineAndAdd_v1, nameof(StreamWriterWriteLineAndAdd_v1)),
	new(StreamWriterWriteLineAndAdd_v2, nameof(StreamWriterWriteLineAndAdd_v2))
];

List<int> actualResults = new();

var semaphore = new SemaphoreSlim(initialCount: 0);

s_Writer = new StreamWriter("output.txt");
foreach (var test in tests) {
	try {
		s_Counter = 0;

		Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: Main: Running test '{test.Name}' ...");

		var t1 = new Thread(() => {
			semaphore.Wait();
			test.Action();
			semaphore.Release();
		});
		t1.Start();

		var t2 = new Thread(() => {
			semaphore.Wait();
			test.Action();
			semaphore.Release();
		});
		t2.Start();

		Thread.Sleep(500);

		semaphore.Release(2);

		Thread.Sleep(500);

		semaphore.Wait();
		semaphore.Wait();
	} catch (Exception ex) {
        Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: Exception caught in Main: {ex}");
    }

	Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: Main: Test '{test.Name}' result: Expected == {Iterations * 2}, Actual == {s_Counter} ...");
    Console.Write("Press ENTER to run next test ... ");
	Console.ReadLine();
    Console.WriteLine();

    actualResults.Add(s_Counter);
}

for (int i = 0; i < actualResults.Count; i++) {
	var test = tests[i];
	Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: Main: Test '{test.Name}' result: Expected == {Iterations * 2}, Actual == {actualResults[i]} ...");
}

Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: Main: Closing StreamWriter ...");
s_Writer.Close();

void AddOnly() {
	for (int i = 0; i < Iterations; i++) {
		s_Counter++;
	}
}

void ConsoleWriteLineAndAdd() {
	for (int i = 0; i < Iterations; i++) {
		Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: s_Counter == {s_Counter}");
		s_Counter++;
	}
}

void StreamWriterWriteLineAndAdd_v1() {
	for (int i = 0; i < Iterations; i++) {
		s_Writer.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: s_Counter == {s_Counter}");
		s_Counter++;
	}
}

void StreamWriterWriteLineAndAdd_v2() {
	try {
		for (int i = 0; i < Iterations; i++) {
			s_Writer.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: s_Counter == {s_Counter}");
			s_Counter++;
		}
	} catch (Exception ex) {
		Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: Exception caught: {ex}");
	}
}

partial class Program {
	public static int s_Counter;

	public static StreamWriter s_Writer;
}

readonly record struct Test(Action Action, string Name);